/*
 * Copyright The University of Chicago
 *
 * All Rights Reserved.
 */

/**
 * @file common/globus_google_drive_common.h Common functions
 */

/**
 * @brief Common Functions
 * @defgroup globus_google_drive_common Common Functions
 * @details
 *     This module provides functions that are used by other modules
 *     within the DSI to provide configuration file parsing, logging, and error
 *     types.
 * @{
 */
#ifndef GLOBUS_GOOGLE_DRIVE_COMMON_H
#define GLOBUS_GOOGLE_DRIVE_COMMON_H 1

#include <stdbool.h>

#include "globus_common.h"
#include "globus_gridftp_server.h"
#include "globus_dsi_rest.h"
#include "jansson.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
#ifndef json_array_foreach
#define json_array_foreach(array, index, value) \
        for(index = 0; \
                index < json_array_size(array) && (value = json_array_get(array, index)); \
                index++)
#endif
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

extern globus_module_descriptor_t globus_i_google_drive_common_module;
/**
 * @brief Module descriptor
 */
#define GLOBUS_GOOGLE_DRIVE_COMMON_MODULE (&globus_i_google_drive_common_module)
#define _GDSL(s) globus_common_i18n_get_string(\
     GLOBUS_GOOGLE_DRIVE_COMMON_MODULE, \
     s)

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
GlobusDebugDeclare(GLOBUS_GOOGLE_DRIVE);
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

/**
 * @brief Configuration key parsing state
 * @details
 *     This type is used to represent a parameter in the DSI configuration
 *     file.
 */
typedef
struct globus_i_google_drive_config_key_s
{
    /**
     * Type of value that this configuration key may hold.
     */
    enum
    {
        /**
         * This is a string parameter. The value pointed to by
         * *value_string* is updated when parsing this configuration option
         */
        GLOBUS_GOOGLE_DRIVE_CONFIG_STRING,
        /**
         * This is an integer parameter. The value pointed to by
         * *value_int* is updated when parsing this configuration option
         */
        GLOBUS_GOOGLE_DRIVE_CONFIG_INT,
        /**
         * This is an off_t parameter. The value pointed to by
         * *value_off_t* is updated when parsing this configuration option
         */
        GLOBUS_GOOGLE_DRIVE_CONFIG_OFF_T,
    }                                   value_type;
    /**
     * The name of the configuration parameter.
     */
    const char                         *key_name;
    /**
     * A pointer to a string that is updated with a copy of the value
     * of the parameter, if the *value_type* enumeration value is
     * `GLOBUS_GOOGLE_DRIVE_CONFIG_STRING`. Otherwise this is ignored.
     */
    char                              **value_string;
    /**
     * A pointer to an integer that is updated with the value
     * of the parameter, if the *value_type* enumeration value is
     * `GLOBUS_GOOGLE_DRIVE_CONFIG_INT`. Otherwise this is ignored.
     */
    int                                *value_int;
    /**
     * A pointer to an off_t that is updated with the value
     * of the parameter, if the *value_type* enumeration value is
     * `GLOBUS_GOOGLE_DRIVE_CONFIG_OFF_T`. Otherwise this is ignored.
     */
    off_t                              *value_off_t;
}
globus_i_google_drive_config_key_t;


globus_result_t
globus_i_google_drive_config_parse_file(
    const char                         *config_path,
    globus_i_google_drive_config_key_t *config_keys,
    bool                                ignore_extra);

globus_result_t
globus_i_google_drive_config_parse_key(
    const char                         *config_path,
    const int                           lineno,
    const char                         *key,
    const char                         *val,
    globus_i_google_drive_config_key_t *config_keys,
    bool                               *parsed_option);

globus_result_t
globus_i_google_drive_config_parse_string(
    const char                         *config_string,
    globus_i_google_drive_config_key_t *config_keys,
    bool                                ignore_extra);

globus_result_t
globus_google_drive_evaluate_path(
    const char                         *orig_path,
    char                              **updated_pathp);


/**
 * @brief File Metadata Model
 * @ingroup globus_gridftp_server_google_drive
 * @details
 *    This data type is used to model the file metadata available on Google
 *    Drive, and is passed in to the Drive API and out to the cache. The
 *    data are almost all pointers, so that NULL values can be used to
 *    prevent overwriting metadata we don't specify in our requests.
 */
typedef
struct globus_google_drive_file_meta_s
{
    /** The ID of the file */
    char                               *id;
    /** The name of the file */
    char                               *name;
    /**
     * The object id of the team drive where this file lives. This may be
     * NULL if the file is not owned by a team drive.
     */
    char                               *driveId;
    /** The mime type of the file */
    char                               *mimeType;
    /**
     * The time the file was created. This is only used when listing
     * resumable upload state files to decide which ones to delete.
     */
    time_t                              createdTime_value;
    /**
     * True if createdTime_value contains a valid value.
     */
    bool                                createdTime_set;
    /**
     * The last time the current user viewed the file. This is used as
     * the atime in the stat result.
     */
    time_t                              viewedByMeTime_value;
    /**
     * True if viewedByMeTime_value contains a valid value.
     */
    bool                                viewedByMeTime_set;
    /**
     * The last time anyone modified the file. This is used as
     * the mtime in the stat result.
     */
    time_t                              modifiedTime_value;
    /**
     * True if modifiedTime_value contains a valid value.
     */
    bool                                modifiedTime_set;
    /**
     * True if the current user is the owner of the file.
     */
    bool                                ownedByMe_value;
    /**
     * True if ownedByMe_value contains a valid value.
     */
    bool                                ownedByMe_set;
    /**
     * The number of elements in the ownerEmailAddress array.
     */
    size_t                              num_ownerEmailAddress;
    /**
     * An array of email addresses of owners of the file.
     */
    char                              **ownerEmailAddress;
    /**
     * True if the current user has permissions to modify the file.
     */
    bool                                canEdit_value;
    /**
     * True if canEdit_value contains a valid value.
     */
    bool                                canEdit_set;
    /**
     * True if the current user has starred this file.
     */
    bool                                starred_value;
    /**
     * True if starred_value contains a valid value.
     */
    bool                                starred_set;
    /**
     * True if the file or a parent folder is in the trash
     */
    bool                                trashed_value;
    /**
     * True if trashed_value contains a valid value.
     */
    bool                                trashed_set;
    /**
     * True if the file has been explicitly added to the trash
     */
    bool                                explicitlyTrashed_value;
    /**
     * True if explicitlyTrashed_value contains a valid value.
     */
    bool                                explicitlyTrashed_set;
    /**
     * The md5 checksum of the file (only for files with binary content).
     */
    char                                md5Checksum[33];
    /**
     * The size of the file
     */
    off_t                               size_value;
    /**
     * True if size_value contains a valid value.
     */
    bool                                size_set;
    /**
     * The number of elements in the parents array
     */
    size_t                              num_parents;
    /**
     * An array of IDs of the folders which are parents of this file.
     */
    char                              **parents;
    /**
     * In a changes list response only, this indicates that the file was
     * removed (either deleted or unshared) and can no longer be accessed
     * by the current user.
     */
    bool                                removed;
}
globus_google_metadata_t;

/**
 * @brief Prefix for Google Docs mime types
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_APPS_PREFIX       "application/vnd.google-apps."
/**
 * @brief Google Drive folder mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_FOLDER_TYPE GLOBUS_GOOGLE_APPS_PREFIX"folder"
/**
 * @brief Google Docs document mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_DOCUMENT    GLOBUS_GOOGLE_APPS_PREFIX"document"
/**
 * @brief Google Docs drawing mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_DRAWING     GLOBUS_GOOGLE_APPS_PREFIX"drawing"
/**
 * @brief Google Docs presentation mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_PRESENTATION \
                                        GLOBUS_GOOGLE_APPS_PREFIX"presentation"
/**
 * @brief Google Docs spreadsheet mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_SPREADSHEET GLOBUS_GOOGLE_APPS_PREFIX"spreadsheet"
/**
 * @brief Google Docs map mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_MAP         GLOBUS_GOOGLE_APPS_PREFIX"map"
/**
 * @brief Google Docs form mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_FORM        GLOBUS_GOOGLE_APPS_PREFIX"form"
/**
 * @brief Google Docs site mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_SITE        GLOBUS_GOOGLE_APPS_PREFIX"site"
/**
 * @brief Google Docs jamboard mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_JAM         GLOBUS_GOOGLE_APPS_PREFIX"jam"
/**
 * @brief Google Docs shortcut mime type
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_SHORTCUT    GLOBUS_GOOGLE_APPS_PREFIX"shortcut"

/**
 * @brief Google Docs document extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_DOCUMENT_EXTENSION \
                                        ".gdoc"
/**
 * @brief Google Docs drawing extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_DRAWING_EXTENSION \
                                        ".gdraw"
/**
 * @brief Google Docs presentation extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_PRESENTATION_EXTENSION \
                                        ".gslides"
/**
 * @brief Google Docs spreadsheet extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_SPREADSHEET_EXTENSION \
                                        ".gsheet"
/**
 * @brief Google Docs map extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_MAP_EXTENSION \
                                        ".gmap"
/**
 * @brief Google Docs form extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_FORM_EXTENSION \
                                        ".gform"
/**
 * @brief Google Docs site extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_SITE_EXTENSION \
                                        ".gsite"
/**
 * @brief Google Docs jamboard extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_JAM_EXTENSION \
                                        ".gjam"
/**
 * @brief Google Docs shortcut extension
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_SHORTCUT_EXTENSION \
                                        ".gshortcut"
const char *
globus_google_drive_doc_extension(
    const char                         *mime_type);

const char *
globus_google_drive_doc_mime_type(
    const char                          *name);

bool
globus_google_drive_mime_type_is_google_docs(
    const char                         *mime_type);

/**
 * @brief Name of the *Trash* pseudo folder
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_NAME_TRASH  "Trash"
/**
 * @brief Name of the *Starred* pseudo folder
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_NAME_STARRED \
                                        "Starred"
/**
 * @brief Name of the *My Drive* pseudo folder
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_NAME_MY_DRIVE \
                                        "My Drive"
/**
 * @brief Name of the *Team Drives* pseudo folder
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_NAME_TEAM_DRIVES \
                                        "Team Drives"
/**
 * @brief Name of the *Shared With Me* pseudo folder
 * @hideinitializer
 */
#define GLOBUS_GOOGLE_DRIVE_NAME_SHARED_WITH_ME \
                                        "Shared With Me"

globus_result_t
globus_google_drive_name_translate_in(
    const char                         *input_name,
    char                              **output_namep,
    const char                        **output_mime_typep);

globus_result_t
globus_google_drive_name_translate_out(
    const globus_google_metadata_t     *input_metadata,
    char                              **output_namep);

globus_result_t
globus_google_metadata_parse(
    json_t                             *json,
    globus_google_metadata_t           *file_meta);

globus_result_t
globus_google_drive_metadata_parse(
    json_t                             *json,
    globus_google_metadata_t           *file_meta);

globus_result_t
globus_google_metadata_copy(
    globus_google_metadata_t           *dest,
    const globus_google_metadata_t     *source);

globus_result_t
globus_google_metadata_convert(
    const globus_google_metadata_t     *file_meta,
    json_t                            **jsonp);

void
globus_google_metadata_destroy(
    globus_google_metadata_t           *file_meta);

void
globus_google_metadata_destroy_void(
    void                               *file_meta);

/**
 * @brief GridFTP Path Resolution State
 * @ingroup globus_google_drive_common
 * @details
 *     This structure is used when resolving a path to an array of
 *     Google Drive Object IDs.
 */
typedef
struct globus_google_drive_path_s
{
    /** The number of path components (separated by '/') in the path */
    size_t                              count;
    /**
     * An array of path components representing the
     * names, object ids, and mime types (if known) of path being resolved
     */
    struct globus_google_drive_path_component_s
    {
        /** The translate_in name of the path component. */
        char                           *name;
        /** The id of the object if known */
        char                           *object_id;
        /** The mime type of the object if known (based on the name) */
        const char                     *mime_type;
    }
    *components;
}
globus_google_drive_path_t;

globus_result_t
globus_google_drive_path_split(
    const char                         *path,
    globus_google_drive_path_t         *split_path);

globus_result_t
globus_google_drive_partial_path_construct(
    const globus_google_drive_path_t   *path,
    size_t                              name_count,
    char                              **partial_pathp);
/**
 * @brief Debugging Log Levels
 */
typedef enum
{
    /** Used to log function entry/exit. */
    GLOBUS_GOOGLE_DRIVE_TRACE         = 1<<0,
    /** Used to log function execution details. */
    GLOBUS_GOOGLE_DRIVE_DEBUG         = 1<<1,
    /** Used to log high level operations. */
    GLOBUS_GOOGLE_DRIVE_INFO          = 1<<2,
    /** Used to log warnings. */
    GLOBUS_GOOGLE_DRIVE_WARN          = 1<<3,
    /** Used to log errors. */
    GLOBUS_GOOGLE_DRIVE_ERROR         = 1<<4,
    /** Used to enable all log levels. */
    GLOBUS_GOOGLE_DRIVE_ALL           = (1<<5)-1
}
globus_google_drive_debug_level_t;

extern const char * globus_i_google_drive_debug_level_names[];

extern const globus_object_type_t
        GLOBUS_GOOGLE_DRIVE_ERROR_TYPE_DEFINITION;
#define GLOBUS_GOOGLE_DRIVE_ERROR_TYPE \
        (&GLOBUS_GOOGLE_DRIVE_ERROR_TYPE_DEFINITION)

void
globus_google_drive_error_add_info(
    globus_object_t *                   error,
    globus_dsi_rest_key_array_t        *info);

const char *
globus_google_drive_error_get_info(
    globus_object_t                    *error,
    const char                         *key);

#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE_CODE_NAME \
        "GD-Response-Code"

#define GLOBUS_GOOGLE_DRIVE_ERROR_RESPONSE_GOOGLE_DRIVE \
        "Internal error (google drive)."
#define GLOBUS_GOOGLE_DRIVE_ERROR_RESPONSE_OAUTH \
        "Invalid security token."
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_METHOD(method) \
        {.key = "GD-Method", .value = (method)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_URI(uri) \
        {.key = "GD-URI", .value = (uri)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_OPERATION(operation) \
        {.key = "GD-Operation", .value = (operation)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_JSON(json) \
        {.key = "GD-JSON", .value = (json)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_ERROR(e) \
        {.key = "GD-Error", .value = (e)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_ERROR_DESCRIPTION(d) \
        {.key = "GD-Error-Description", .value = (d)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_DETAILS(details) \
        {.key = "GD-Details", .value = (details)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_PATH(p) \
        {.key = "GD-Path", .value = (p)}
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_PARSE_ERROR(error) \
        {.key = "GD-Parse-Error", .value = (error) }
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE(r) \
        {.key = "GD-Response", .value = (r) }
#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE_CODE(c) \
        {.key = GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE_CODE_NAME, .value = c}

#define GLOBUS_GOOGLE_DRIVE_ERROR_FIELDS(...) \
        (&(globus_dsi_rest_key_array_t) \
        { \
           .count = (sizeof((globus_dsi_rest_key_value_t[]){__VA_ARGS__}) / sizeof(globus_dsi_rest_key_value_t)), \
           .key_value = (globus_dsi_rest_key_value_t[]) \
           { \
               __VA_ARGS__ \
           } \
        })
#define GLOBUS_GOOGLE_DRIVE_KEYVALUE_ARRAY(...) \
        ((globus_dsi_rest_key_array_t) \
        { \
           .count = (sizeof((globus_dsi_rest_key_value_t[]){__VA_ARGS__}) / sizeof(globus_dsi_rest_key_value_t)), \
           .key_value = (globus_dsi_rest_key_value_t[]) \
           { \
               __VA_ARGS__ \
           } \
        })

/**
 * @brief Error result parsing a json object with error details
 * @hideinitializer
 */
#define GlobusGoogleDriveErrorJson(j) \
    globus_error_put(GlobusGoogleDriveErrorObjJson(j))

#define GlobusGoogleDriveErrorObjJson(j) \
        GlobusGFSErrorObjInternalError( \
            GlobusGoogleDriveErrorObjGoogleDrive( \
                NULL, \
                GLOBUS_GOOGLE_DRIVE_ERROR_FIELDS( \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_PARSE_ERROR( \
                                (j)->text))), NULL)

#define GlobusGoogleDriveErrorInvalidDocument(path) \
    globus_error_put(GlobusGoogleDriveErrorInvalidDocumentObj(path))
#define GlobusGoogleDriveErrorInvalidDocumentObj(path) \
        GlobusGFSErrorObjInvalidPathName( \
            GlobusGoogleDriveErrorObjGoogleDrive(NULL, NULL), (path))

/**
 * @brief Error constructor for a configuration error
 * @hideinitializer
 */
#define GlobusGoogleDriveErrorConfig(operation, path, details) \
    globus_error_put( \
        GlobusGoogleDriveErrorObjConfig(operation, path, details))

#define GlobusGoogleDriveErrorObjConfig(operation, path, details) \
    GlobusGFSErrorObjConfigurationError( \
            GlobusGoogleDriveErrorObjGoogleDrive( \
                NULL, \
                GLOBUS_GOOGLE_DRIVE_ERROR_FIELDS( \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_PATH(path), \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_OPERATION(operation), \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_DETAILS(details)) \
                ))

#define GlobusGoogleDriveErrorAmbiguousPath(p) \
    globus_error_put( \
        GlobusGoogleDriveErrorObjAmbiguousPath(p))

#define GlobusGoogleDriveErrorObjAmbiguousPath(p) \
        GlobusGFSErrorObjAmbiguousPath(NULL, (p))

#define GlobusGoogleDriveErrorOAuthResponse( \
        method, uri, code, error, description) \
    globus_error_put(GlobusGoogleDriveErrorObjOAuthResponse( \
            method, uri, code, error, description))
#define GlobusGoogleDriveErrorObjOAuthResponse( \
    method, uri, code, error, description) \
        GlobusGFSErrorObjInternalError( \
        GlobusGoogleDriveErrorObjGoogleDrive( \
                NULL, \
                GLOBUS_GOOGLE_DRIVE_ERROR_FIELDS( \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_METHOD(method), \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_URI(uri), \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE_CODE(code), \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_ERROR(error), \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_ERROR_DESCRIPTION(description),\
                    )), NULL)
#define GlobusGoogleDriveErrorSecurityToken(operation, path, details) \
    globus_error_put( \
            GlobusGoogleDriveErrorObjSecurityToken( \
                    operation, path, details))
#define GlobusGoogleDriveErrorObjSecurityToken(operation, path, details) \
        GlobusGFSErrorObjInternalError( \
                GlobusGoogleDriveErrorObjGoogleDrive( \
                    NULL, \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELDS( \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_DETAILS(operation), \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_DETAILS(details), \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_PATH(path), \
                        )), NULL)
/**
 * @brief Error result constructor for GLOBUS_GOOGLE_DRIVE_ERROR_JSON_RESPONSE
 * @hideinitializer
 */
#define GlobusGoogleDriveErrorJsonResponse(method, uri, hr, je) \
    globus_error_put(GlobusGoogleDriveErrorObjJsonResponse(method, uri, hr, je))
#define GlobusGoogleDriveErrorObjJsonResponse(method, uri, hr, je) \
        GlobusGFSErrorObjInternalError( \
                GlobusGoogleDriveErrorObjGoogleDrive( \
                    NULL, \
                    GLOBUS_GOOGLE_DRIVE_ERROR_FIELDS( \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_METHOD(method), \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_URI(uri), \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE_CODE(hr), \
                        GLOBUS_GOOGLE_DRIVE_ERROR_FIELD_RESPONSE(je))), NULL)

#define GlobusGoogleDriveErrorParameterNotImplemented(info) \
        globus_error_put(GlobusGoogleDriveErrorObjParameterNotImplemented( \
                info))
#define GlobusGoogleDriveErrorObjParameterNotImplemented(info) \
        GlobusGFSErrorObjNotImplemented( \
            GlobusGoogleDriveErrorObjGoogleDrive(NULL, info))

#define GlobusGoogleDriveErrorData() \
    globus_error_put(GlobusGoogleDriveErrorObjData())

#define GlobusGoogleDriveErrorObjData() \
    GlobusGFSErrorObjDataChannelCommunicationFailure(NULL)

#define GlobusGoogleDriveErrorGoogleDrive(cause, info) \
        globus_error_put(GlobusGoogleDriveErrorObjGoogleDrive( \
            (cause), \
            (info)))
#define GlobusGoogleDriveErrorObjGoogleDrive( \
        cause, info) \
        globus_google_drive_error_construct( \
            NULL, \
            (cause), \
            (info))
/**
 * @brief Log a message at the specified logging level
 * @hideinitializer
 * @param[in] level
 *     A globus_google_drive_debug_level_t value.
 * @param[in] ...
 *     Printf string for the log message
 */
#define GlobusGoogleDriveLog(level, ...) \
    do \
    { \
        int level__ = level; \
        if (level__ > GLOBUS_GOOGLE_DRIVE_ERROR \
            || level__  < 0 \
            || globus_i_google_drive_debug_level_names[level__] == NULL) \
        { \
            level__ = 1; \
        } \
        if (GlobusDebugTrue(GLOBUS_GOOGLE_DRIVE, level__)) \
        { \
            flockfile(GlobusDebugMyFile(GLOBUS_GOOGLE_DRIVE)); \
            GlobusDebugPrintf(GLOBUS_GOOGLE_DRIVE, level__, \
                ("google_drive: %5s: %"PRIiMAX": %s: ", \
                 globus_i_google_drive_debug_level_names[level__], \
                 (intmax_t) getpid(), __func__)); \
            GlobusDebugMyPrintf(GLOBUS_GOOGLE_DRIVE, \
                (__VA_ARGS__)); \
            funlockfile(GlobusDebugMyFile(GLOBUS_GOOGLE_DRIVE)); \
        } \
    } while (0)

globus_object_t *
globus_google_drive_error_construct(
    globus_module_descriptor_t *        base_source,
    globus_object_t *                   base_cause,
    globus_dsi_rest_key_array_t        *error_info);

globus_object_t *
globus_google_drive_error_initialize(
    globus_object_t *                   error,
    globus_module_descriptor_t *        base_source,
    globus_object_t *                   base_cause,
    globus_dsi_rest_key_array_t        *error_info);


/**
 * @brief Log a message at the GLOBUS_GOOGLE_DRIVE_TRACE level.
 * @hideinitializer
 */
#define GlobusGoogleDriveTrace(...)   GlobusGoogleDriveLog(GLOBUS_GOOGLE_DRIVE_TRACE, __VA_ARGS__)
/**
 * @brief Log a message at the GLOBUS_GOOGLE_DRIVE_INFO level.
 * @hideinitializer
 */
#define GlobusGoogleDriveInfo(...)    GlobusGoogleDriveLog(GLOBUS_GOOGLE_DRIVE_INFO,  __VA_ARGS__)
/**
 * @brief Log a message at the GLOBUS_GOOGLE_DRIVE_DEBUG level.
 * @hideinitializer
 */
#define GlobusGoogleDriveDebug(...)   GlobusGoogleDriveLog(GLOBUS_GOOGLE_DRIVE_DEBUG, __VA_ARGS__)
/**
 * @brief Log a message at the GLOBUS_GOOGLE_DRIVE_WARN level.
 * @hideinitializer
 */
#define GlobusGoogleDriveWarning(...) GlobusGoogleDriveLog(GLOBUS_GOOGLE_DRIVE_WARN, __VA_ARGS__)
/**
 * @brief Log a message at the GLOBUS_GOOGLE_DRIVE_ERROR level.
 * @hideinitializer
 */
#define GlobusGoogleDriveError(...)   GlobusGoogleDriveLog(GLOBUS_GOOGLE_DRIVE_ERROR, __VA_ARGS__)

/**
 * @brief Log function entry
 * @hideinitializer
 */
#define GlobusGoogleDriveEnter() GlobusGoogleDriveTrace("enter\n")
/**
 * @brief Log function exit with an integer result.
 * @hideinitializer
 */
#define GlobusGoogleDriveExitInt(i) GlobusGoogleDriveTrace("exit: %d\n", (i))
/**
 * @brief Log function exit with a boolean result.
 * @hideinitializer
 */
#define GlobusGoogleDriveExitBool(b) GlobusGoogleDriveTrace("exit: %s\n", (b) ? "true" : "false")
/**
 * @brief Log function exit with a string result.
 * @hideinitializer
 */
#define GlobusGoogleDriveExitString(s) GlobusGoogleDriveTrace("exit: %s\n", (s) != NULL ? (s) : "NULL")
/**
 * @brief Log function exit.
 * @hideinitializer
 */
#define GlobusGoogleDriveExit() GlobusGoogleDriveTrace("exit\n")

/**
 * @brief Log an error object.
 * @hideinitializer
 * @param[in] level
 *     The debugging level to log the message at.
 * @param[in] result
 *     The result to log.
 */
#define GlobusGoogleDriveLogResult(level, result) \
        do { \
            if (result) \
            { \
                char *errstr = globus_error_print_chain(globus_error_peek(result)); \
                GlobusGoogleDriveLog(level, "result=%#x message=%s\n", result, errstr); \
                free(errstr); \
            } \
            else \
            { \
                GlobusGoogleDriveLog(level, "result=GLOBUS_SUCCESS\n"); \
            } \
        } while (0)
/**
 * @brief Log function exit with a boolean result.
 * @hideinitializer
 */
#define GlobusGoogleDriveExitResult(result) \
        do { \
            if (result) \
            { \
                char *errstr = globus_error_print_chain(globus_error_peek(result)); \
                GlobusGoogleDriveError("exit: result=%#x message=%s\n", result, errstr); \
                free(errstr); \
            } \
            else \
            { \
                GlobusGoogleDriveTrace("exit: result=GLOBUS_SUCCESS\n"); \
            } \
        } while (0)
#ifdef __cplusplus
}
#endif

/**
 * @}
 */
#endif /* GLOBUS_GOOGLE_DRIVE_COMMON_H */
